@@ -10,7 +10,6 @@ from .rewrite_import_typing import RewriteImportTyping
 from .rewrite_import import RewriteImport
 
 import pluthon as plt
-from uplc.uplc_ast import BuiltInFun
 
 STATEMONAD = "s"
 
@@ -18,9 +17,7 @@ STATEMONAD = "s"
 BinOpMap = {
     Add: {
         IntegerType: {
-            IntegerType: {
-                plt.AddInteger,
-            }
+            IntegerType: plt.AddInteger,
         },
         ByteStringType: {
             ByteStringType: plt.AppendByteString,
@@ -51,30 +48,41 @@ BinOpMap = {
 CmpMap = {
     Eq: {
         IntegerType: {
-            IntegerType: {
-                plt.EqualsInteger,
-            }
+            IntegerType: plt.EqualsInteger,
         },
         ByteStringType: {
-            ByteStringType: {
-                plt.EqualsByteString,
-            }
+            ByteStringType: plt.EqualsByteString,
         },
         BoolType: {
-            BoolType: {
-                plt.EqualsBool,
-            }
+            BoolType: plt.EqualsBool,
         },
     },
     Lt: {
         IntegerType: {
-            IntegerType: {
-                BuiltInFun.LessThanInteger,
-            }
+            IntegerType: plt.LessThanInteger,
         },
     },
 }
 
+TransformExtParamsMap = {
+    IntegerType: lambda x: plt.UnIData(x),
+    StringType: lambda x: plt.DecodeUtf8(plt.UnBData(x)),
+    ByteStringType: lambda x: plt.UnBData(x),
+    ListType: lambda x: plt.UnListData(x),
+    DictType: lambda x: plt.UnMapData(x),
+    UnitType: lambda x: plt.Lambda(["_"], plt.Unit()),
+    BoolType: lambda x: plt.NotEqualsInteger(plt.UnIData(x), plt.Integer(0)),
+}
+TransformOutputMap = {
+    IntegerType: lambda x: plt.IData(x),
+    StringType: lambda x: plt.BData(plt.EncodeUtf8(x)),
+    ByteStringType: lambda x: plt.BData(x),
+    ListType: lambda x: plt.ListData(x),
+    DictType: lambda x: plt.MapData(x),
+    UnitType: lambda x: plt.Lambda(["_"], plt.Unit()),
+    BoolType: lambda x: plt.IData(plt.IfThenElse(x, plt.Integer(1), plt.Integer(0))),
+}
+
 ConstantMap = {
     str: plt.Text,
     bytes: plt.ByteString,
@@ -100,10 +108,10 @@ class PythonBuiltIn(Enum):
     range = plt.Lambda(["f", "limit", STATEMONAD], plt.Range(plt.Var("limit")))
 
 
-INITIAL_STATE = extend_statemonad(
+INITIAL_STATE = plt.FunctionalMapExtend(
+    plt.FunctionalMap(),
     [b.name for b in PythonBuiltIn],
     [b.value for b in PythonBuiltIn],
-    plt.FunctionalMap(),
 )
 
 
@@ -175,20 +183,45 @@ class UPLCCompiler(NodeTransformer):
         )
 
     def visit_Module(self, node: TypedModule) -> plt.AST:
+        # find main function
+        # TODO can use more sophisiticated procedure here i.e. functions marked by comment
+        main_fun = None
+        for s in node.body:
+            if isinstance(s, FunctionDef) and s.name == "validator":
+                main_fun = s
         cp = plt.Program(
             "0.0.1",
             # TODO directly unwrap supposedly int/byte data? how?
-            plt.Let(
-                [
-                    (
-                        "g",
+            TransformOutputMap.get(main_fun.returns, lambda x: x)(
+                plt.Lambda(
+                    [f"p{i}" for i, _ in enumerate(main_fun.args.args)],
+                    plt.Let(
+                        [
+                            ("s", INITIAL_STATE),
+                            (
+                                "g",
+                                plt.FunctionalMapAccess(
+                                    plt.Apply(
+                                        self.visit_sequence(node.body), plt.Var("s")
+                                    ),
+                                    plt.ByteString("validator".encode("utf8")),
+                                    plt.TraceError("NameError"),
+                                ),
+                            ),
+                        ],
                         plt.Apply(
-                            plt.Apply(self.visit_sequence(node.body), INITIAL_STATE),
-                            plt.ByteString("main".encode("utf8")),
+                            plt.Var("g"),
+                            plt.Var("g"),
+                            *[
+                                TransformExtParamsMap.get(a.typ, lambda x: x)(
+                                    plt.Var(f"p{i}")
+                                )
+                                for i, a in enumerate(main_fun.args.args)
+                            ],
+                            plt.Var("s"),
                         ),
-                    )
-                ],
-                plt.Apply(plt.Var("g"), plt.Var("g")),
+                    ),
+                ),
             ),
         )
         return cp
@@ -202,7 +235,7 @@ class UPLCCompiler(NodeTransformer):
         return plt.Lambda([STATEMONAD], plt_type(node.value))
 
     def visit_NoneType(self, _: typing.Optional[typing.Any]) -> plt.AST:
-        return plt.Lambda([STATEMONAD], plt.Unit())
+        return plt.Lambda([STATEMONAD], plt.NoneData())
 
     def visit_Assign(self, node: TypedAssign) -> plt.AST:
         assert (
@@ -215,10 +248,10 @@ class UPLCCompiler(NodeTransformer):
         # (\{STATEMONAD} -> (\x -> if (x ==b {self.visit(node.targets[0])}) then ({compiled_e} {STATEMONAD}) else ({STATEMONAD} x)))
         return plt.Lambda(
             [STATEMONAD],
-            extend_statemonad(
+            plt.FunctionalMapExtend(
+                plt.Var(STATEMONAD),
                 [node.targets[0].id],
                 [plt.Apply(compiled_e, plt.Var(STATEMONAD))],
-                plt.Var(STATEMONAD),
             ),
         )
 
@@ -230,7 +263,7 @@ class UPLCCompiler(NodeTransformer):
                 plt.FunctionalMapAccess(
                     plt.Var(STATEMONAD),
                     plt.ByteString(node.id.encode()),
-                    plt.Trace("NameError", plt.Error()),
+                    plt.TraceError("NameError"),
                 ),
             )
         raise NotImplementedError(f"Context {node.ctx} not supported")
@@ -258,7 +291,10 @@ class UPLCCompiler(NodeTransformer):
                     plt.Var("g"),
                     # pass the function to itself for recursion
                     plt.Var("g"),
+                    # pass in all arguments evaluated with the statemonad
                     *(plt.Apply(self.visit(a), plt.Var(STATEMONAD)) for a in node.args),
+                    # eventually pass in the state monad as well
+                    plt.Var(STATEMONAD),
                 ),
             ),
         )
@@ -267,26 +303,30 @@ class UPLCCompiler(NodeTransformer):
         body = node.body.copy()
         if not isinstance(body[-1], Return):
             tr = Return(None)
-            tr.typ = UnitType
+            tr.typ = NoneType
             assert (
-                node.typ.rettyp == UnitType
+                node.typ.rettyp == NoneType
             ), "Function has no return statement but is supposed to return not-None value"
             body.append(tr)
         compiled_body = self.visit_sequence(body[:-1])
         compiled_return = self.visit(body[-1].value)
-        args_state = extend_statemonad(
+        args_state = plt.FunctionalMapExtend(
+            plt.Var(STATEMONAD),
             # under the name of the function, it can access itself
             [node.name] + [a.arg for a in node.args.args],
             [plt.Var("f")] + [plt.Var(f"p{i}") for i in range(len(node.args.args))],
-            plt.Var(STATEMONAD),
         )
         return plt.Lambda(
             [STATEMONAD],
-            extend_statemonad(
+            plt.FunctionalMapExtend(
+                plt.Var(STATEMONAD),
                 [node.name],
                 [
                     plt.Lambda(
-                        ["f"] + [f"p{i}" for i in range(len(node.args.args))],
+                        # expect the statemonad again -> this is the basis for internally available values
+                        ["f"]
+                        + [f"p{i}" for i in range(len(node.args.args))]
+                        + [STATEMONAD],
                         plt.Apply(
                             compiled_return,
                             plt.Apply(
@@ -296,7 +336,6 @@ class UPLCCompiler(NodeTransformer):
                         ),
                     )
                 ],
-                plt.Var(STATEMONAD),
             ),
         )
 
@@ -387,12 +426,7 @@ class UPLCCompiler(NodeTransformer):
                             plt.Lambda(
                                 ["i", "xs", "f"],
                                 plt.Ite(
-                                    plt.Force(
-                                        plt.Apply(
-                                            plt.BuiltIn(BuiltInFun.NullList),
-                                            plt.Var("xs"),
-                                        )
-                                    ),
+                                    plt.NullList(plt.Var("xs")),
                                     self.visit(
                                         TypedCall(
                                             TypedName("print", ctx=Load()),
@@ -404,29 +438,21 @@ class UPLCCompiler(NodeTransformer):
                                         )
                                     ),
                                     plt.Ite(
-                                        plt.Apply(
-                                            plt.BuiltIn(BuiltInFun.EqualsInteger),
+                                        plt.EqualsInteger(
                                             plt.Var("i"),
                                             plt.Integer(0),
                                         ),
-                                        plt.Force(
-                                            plt.Apply(
-                                                plt.BuiltIn(BuiltInFun.HeadList),
-                                                plt.Var("xs"),
-                                            )
+                                        plt.HeadList(
+                                            plt.Var("xs"),
                                         ),
                                         plt.Apply(
                                             plt.Var("f"),
-                                            plt.Apply(
-                                                plt.BuiltIn(BuiltInFun.SubtractInteger),
+                                            plt.SubtractInteger(
                                                 plt.Var("i"),
                                                 plt.Integer(1),
                                             ),
-                                            plt.Force(
-                                                plt.Apply(
-                                                    plt.BuiltIn(BuiltInFun.TailList),
-                                                    plt.Var("xs"),
-                                                )
+                                            plt.TailList(
+                                                plt.Var("xs"),
                                             ),
                                             plt.Var("f"),
                                         ),
